{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\n# Format MERFISH data\n\nThe following script formats MERFISH data acquired from cultured U2-OS cells.\nThis is a good example of:\n\n* converting multipage TIFF files that contain every round, channel of a FOV in a single file\n* multiple fields of view (FOV)\n* fetching primary images and auxiliary images from the same file\n* assigning physical coordinates from CSV file\n\n<div class=\"alert alert-info\"><h4>Note</h4><p>This example is provided for illustrative purposes, demonstrating how the\n   :py:class:`.TileFetcher` is used in practice. It will need to be adapted to meet\n   the specific needs of your data.</p></div>\n\nThe experiment had 496 of fields of view. Each FOV is stored in one file consisting of primary\nimages (8 rounds, 2 channels, 1 z-tile) and one DAPI-stained image. The tile index that maps to\neach round and channel is defined in the :py:class:`.FetchedTile` subclasses.\n\n## input data structure:\n\n    \u2514\u2500\u2500 parent\n        \u251c\u2500\u2500 stagePos.csv\n        \u251c\u2500\u2500 fov_0.tif\n        \u251c\u2500\u2500 fov_1.tif\n        \u251c\u2500\u2500 fov_2.tif\n        \u251c\u2500\u2500 ...\n\nThe locations of the data files for use with this script can be found in the s3_bucket variable.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import argparse\nimport functools\nimport json\nimport os\nfrom typing import IO, Mapping, Union\n\nimport numpy as np\nimport pandas as pd\nfrom skimage.io import imread\nfrom slicedimage import ImageFormat\n\nfrom starfish.core.util.argparse import FsExistsType\nfrom starfish.experiment.builder import FetchedTile, TileFetcher, write_experiment_json\nfrom starfish.types import Axes, Coordinates, CoordinateValue\n\nSHAPE = {Axes.Y: 2048, Axes.X: 2048}\n\n\n# We use this to cache images across tiles.  In the case of the merfish data set, FOVs are saved\n# together in a single file.  To avoid reopening and decoding the TIFF file, we use a single-element\n# cache that maps between file_path and the decoded multipage TIFF.\n@functools.lru_cache(maxsize=1)\ndef cached_read_fn(file_path):\n    return imread(file_path)\n\n\nclass MERFISHTile(FetchedTile):\n    def __init__(self, file_path, r, ch, coordinates):\n        self.file_path = file_path\n        # how to index tiles from indices into multi-page tiff\n        # key is a tuple of round, chan. val is the index\n        self.map = {(0, 0): 0,\n                    (0, 1): 1,\n                    (1, 0): 3,\n                    (1, 1): 2,\n                    (2, 0): 4,\n                    (2, 1): 5,\n                    (3, 0): 6,\n                    (3, 1): 7,\n                    (4, 0): 9,\n                    (4, 1): 8,\n                    (5, 0): 10,\n                    (5, 1): 11,\n                    (6, 0): 12,\n                    (6, 1): 13,\n                    (7, 0): 15,\n                    (7, 1): 14}\n        self.r = r\n        self.ch = ch\n        self._coordinates = coordinates\n\n    @property\n    def shape(self) -> Mapping[Axes, int]:\n        return SHAPE\n\n    @property\n    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:\n        return self._coordinates\n\n    def tile_data(self) -> IO:\n        return cached_read_fn(self.file_path)[self.map[(self.r, self.ch)], :, :]\n\n\nclass MERFISHAuxTile(FetchedTile):\n    def __init__(self, file_path, coordinates):\n        self.file_path = file_path\n        self.dapi_index = 17\n        self._coordinates = coordinates\n\n    @property\n    def shape(self) -> Mapping[Axes, int]:\n        return SHAPE\n\n    @property\n    def coordinates(self) -> Mapping[Union[str, Coordinates], CoordinateValue]:\n        return self._coordinates\n\n    def tile_data(self) -> np.ndarray:\n        return cached_read_fn(self.file_path)[self.dapi_index, :, :]\n\n\nclass MERFISHTileFetcher(TileFetcher):\n    def __init__(self, input_dir, is_dapi):\n        self.input_dir = input_dir\n        self.is_dapi = is_dapi\n        self.coordinates = self.parse_coordinates(input_dir)\n\n    @staticmethod\n    def parse_coordinates(input_dir):\n        filename = os.path.join(input_dir, \"stagePos.csv\")\n        data = pd.read_csv(filename, names=['y_min', 'x_min'])\n        data['x_max'] = data['x_min'] + 200\n        data['y_max'] = data['y_min'] + 200\n        return data\n\n    def make_coordinates(self, fov):\n        return {\n            Coordinates.X: (\n                float(self.coordinates.loc[fov, 'x_min']),\n                float(self.coordinates.loc[fov, 'x_max'])\n            ),\n            Coordinates.Y: (\n                float(self.coordinates.loc[fov, 'y_min']),\n                float(self.coordinates.loc[fov, 'y_max'])\n            ),\n            Coordinates.Z: (0.0, 0.001)\n        }\n\n    def get_tile(\n            self, fov_id: int, round_label: int, ch_label: int, zplane_label: int) -> FetchedTile:\n        filename = os.path.join(self.input_dir, 'fov_{}.tif'.format(fov_id))\n        file_path = os.path.join(self.input_dir, filename)\n        if self.is_dapi:\n            return MERFISHAuxTile(file_path, self.make_coordinates(fov_id))\n        else:\n            return MERFISHTile(file_path, round_label, ch_label, self.make_coordinates(fov_id))\n\n\ndef format_data(input_dir, output_dir):\n\n    input_dir = os.path.abspath(input_dir)\n    output_dir = os.path.abspath(output_dir)\n\n    def add_scale_factors(experiment_json_doc):\n        filename = os.path.join(input_dir, \"scale_factors.json\")\n        with open(filename, 'r') as f:\n            data = json.load(f)\n        experiment_json_doc['extras'] = {\"scale_factors\": data}\n        return experiment_json_doc\n\n    num_fovs = 496\n\n    primary_image_dimensions = {\n        Axes.ROUND: 8,\n        Axes.CH: 2,\n        Axes.ZPLANE: 1,\n    }\n\n    aux_name_to_dimensions = {\n        'nuclei': {\n            Axes.ROUND: 1,\n            Axes.CH: 1,\n            Axes.ZPLANE: 1\n        }\n    }\n\n    write_experiment_json(output_dir,\n                          num_fovs,\n                          tile_format=ImageFormat.TIFF,\n                          primary_image_dimensions=primary_image_dimensions,\n                          aux_name_to_dimensions=aux_name_to_dimensions,\n                          primary_tile_fetcher=MERFISHTileFetcher(input_dir, is_dapi=False),\n                          aux_tile_fetcher={\n                              'nuclei': MERFISHTileFetcher(input_dir, is_dapi=True),\n                          },\n                          postprocess_func=add_scale_factors,\n                          default_shape=SHAPE\n                          )\n\n\nif __name__ == \"__main__\":\n    s3_bucket = \"s3://czi.starfish.data.public/browse/raw/20180820/merfish_u2os/\"\n    input_help_msg = \"Path to raw data. Raw data can be downloaded from: {}\".format(s3_bucket)\n    output_help_msg = \"Path to output experment.json and all formatted images it references\"\n    parser = argparse.ArgumentParser()\n    parser.add_argument(\"input_dir\", type=FsExistsType(), help=input_help_msg)\n    parser.add_argument(\"output_dir\", type=FsExistsType(), help=output_help_msg)\n\n    args = parser.parse_args()\n    format_data(args.input_dir, args.output_dir)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.12"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}